"""Module to monkey patch tree_sitter Nodes to add useful functionality. As these
are extension modules/types, they cannot be extended by inheritance!"""
from typing import List, Tuple

from tree_sitter import Node

from castep_linter.fortran.type_checking import node_type_check


def get_child_by_name(node, typename: str) -> Node:
    """Return the first child node with the requested type"""
    for c in node.named_children:
        if c.type == typename:
            return c

    err = f'"{typename}" not found in children of node {node}'
    raise KeyError(err)


def get_children_by_name(node, typename: str) -> List[Node]:
    """Return all the children with the requested type"""
    return [c for c in node.named_children if c.type == typename]


def get_child_property(node, prop: str) -> Node:
    """Finds a nested property of the form node.subroutine_statment.name"""
    properties = prop.split(".")
    cursor = node
    for p in properties:
        cursor = get_child_by_name(cursor, p)
    return cursor


def split_relational_node(node: Node) -> Tuple[Node, Node]:
    """Split a relational node with a left and right part into the two child nodes"""
    left = node.child_by_field_name("left")

    if left is None:
        err = f"Unable to find left part of node pair: {get_code(node)}"
        raise KeyError(err)

    right = node.child_by_field_name("right")

    if right is None:
        err = f"Unable to find right part of node pair: {get_code(node)}"
        raise KeyError(err)

    return left, right


def get_code(node: Node, *, lower=False) -> str:
    """Return a string of all the text in a node as unicode"""
    if lower:
        return node.text.decode().lower()
    else:
        return node.text.decode()


@node_type_check("string_literal")
def parse_string_literal(node: Node) -> str:
    "Parse a string literal object to get the string"
    return node.raw().strip("\"'")


def add_extra_node_methods():
    """Add extra methods to node class"""
    Node.raw = get_code
    Node.get_child_by_name = get_child_by_name
    Node.get_children_by_name = get_children_by_name
    Node.get_child_property = get_child_property
    Node.split = split_relational_node
    Node.parse_string_literal = parse_string_literal


add_extra_node_methods()



"""Test that all real values are specified by real(kind=dp)"""
from tree_sitter import Node

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran import ArgType, FType, VariableDeclaration, parser


@parser.node_type_check("variable_declaration")
def test_real_dp_declaration(node: Node, error_log: ErrorLogger) -> None:
    """Test that all real values are specified by real(kind=dp)"""

    var_decl = VariableDeclaration(node)

    if var_decl.type in [FType.REAL, FType.COMPLEX]:
        try:
            arg_type, arg_value = var_decl.get_arg(position=1, keyword="kind")
        except KeyError:
            error_log.add_msg("Error", node, "No kind specifier")
            return

        if arg_value.type == "number_literal":
            error_log.add_msg("Error", arg_value, "Numeric kind specifier")

        elif arg_value.type == "identifier" and arg_value.raw(lower=True) != "dp":
            error_log.add_msg("Warning", arg_value, "Invalid kind specifier")

        elif arg_type is ArgType.POSITION:
            error_log.add_msg("Info", arg_value, "Kind specified without keyword")



"""Test that a call of complex(x) has a dp"""
from tree_sitter import Node

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran import CallExpression, parser


@parser.node_type_check("call_expression")
def test_complex_has_dp(node: Node, error_log: ErrorLogger) -> None:
    """Test that a call of complex(x) has a dp"""

    call_expr = CallExpression(node)

    if call_expr.name == "cmplx":
        try:
            _, arg_value = call_expr.get_arg(position=3, keyword="kind")
        except KeyError:
            error_log.add_msg("Error", node, "No kind specifier in complex intrinsic")
            return

        if arg_value.raw(lower=True) != "dp":
            error_log.add_msg("Error", node, "Invalid kind specifier in complex intrinsic")



"""Test that allocate stat is used and checked"""
from tree_sitter import Node

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran import CallExpression, parser


@parser.node_type_check("call_expression")
def test_allocate_has_stat(node: Node, error_log: ErrorLogger) -> None:
    """Test that allocate stat is used and checked"""

    routine = CallExpression(node)

    # Check this is actually an allocate statement
    if routine.name != "allocate":
        return

    # First get the stat variable for this allocate statement
    try:
        _, stat_variable = routine.get_arg(keyword="stat")
    except KeyError:
        err = "No stat on allocate statement"
        error_log.add_msg("Warning", node, err)
        return

    # Find the next non-comment line
    next_node = node.next_named_sibling
    while next_node and next_node.type == "comment":
        next_node = next_node.next_named_sibling

    # Check if that uses the stat variable
    if next_node and next_node.type == "if_statement":
        try:
            relational_expr = next_node.get_child_by_name("parenthesized_expression").get_child_by_name(
                "relational_expression"
            )

        except KeyError:
            error_log.add_msg("Error", stat_variable, "Allocate status not checked")
            return

        lhs, rhs = relational_expr.split()

        if lhs.type == "identifier" and lhs.text.lower() == stat_variable.text.lower():
            return

        if rhs.type == "identifier" and rhs.text.lower() == stat_variable.text.lower():
            return

    error_log.add_msg("Error", stat_variable, "Allocate status not checked")



# pylint: disable=W0621,C0116,C0114
import pytest

from castep_linter import tests
from castep_linter.fortran.parser import get_fortran_parser
from castep_linter.scan_files import run_tests_on_code


@pytest.fixture
def test_list():
    return {"variable_declaration": [tests.test_real_dp_declaration]}


@pytest.fixture
def parser():
    return get_fortran_parser()


def test_real_dp_correct(parser, test_list):
    code = b"real(kind=dp) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 0


@pytest.mark.skip(reason="Not current implemented")
def test_real_dp_correctb(parser, test_list):
    code = b"real, kind(dp) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_real_dp_by_position(parser, test_list):
    code = b"real(dp) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_dp_no_kind(parser, test_list):
    code = b"real, intent(in) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_integer_kind_with_keyword(parser, test_list):
    code = b"real(kind=8) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_integer_kind_without_keyword(parser, test_list):
    code = b"real(8) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_bad_var_kind_without_keyword(parser, test_list):
    code = b"real(x) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_bad_var_kind_with_keyword(parser, test_list):
    code = b"real(kind=x) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_with_other_keyword(parser, test_list):
    code = b"real(lemon=x) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_integer_declaration(parser, test_list):
    code = b"integer :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_derived_type_declaration(parser, test_list):
    code = b"type(z) :: y"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 0



# pylint: disable=W0621,C0116,C0114
import pytest

from castep_linter import tests
from castep_linter.fortran.parser import get_fortran_parser
from castep_linter.scan_files import run_tests_on_code


@pytest.fixture
def test_list():
    return {"subroutine": [tests.test_trace_entry_exit], "function": [tests.test_trace_entry_exit]}


@pytest.fixture
def parser():
    return get_fortran_parser()


def subroutine_wrapper(code):
    return (
        b"""module foo
        subroutine x(y)
        """
        + code
        + b"""
        end subroutine x
        end module foo"""
    )


def test_trace_entry_exit_correct(parser, test_list):
    code = b"""
    call trace_entry("x", stat)
    call trace_exit("x", stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_extra(parser, test_list):
    code = b"""
    call trace_entry("x", stat)
    call bleh()
    call trace_exit("x", stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_keyword(parser, test_list):
    code = b"""
    call trace_entry(string="x", status=stat)
    call trace_exit(string="x", status=stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_by_param(parser, test_list):
    code = b"""
    character(len=100), parameter :: sub_name = "x"
    call trace_entry(sub_name, stat)
    call trace_exit(sub_name, stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_by_param_extra(parser, test_list):
    code = b"""
    character(len=100), parameter :: sub_name = "x"
    character(len=100), parameter :: bleh = othervar
    integer :: p
    type(myvar) :: z
    call trace_entry(sub_name, stat)
    call trace_exit(sub_name, stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_missing(parser, test_list):
    code = b"""
    call trace_exit("x", stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_trace_exit_missing(parser, test_list):
    code = b"""
    call trace_entry("x", stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_trace_entry_exit_missing(parser, test_list):
    code = b""
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_wrong_name(parser, test_list):
    code = b"""
    call trace_entry("y", stat)
    call trace_exit("y", stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_wrong_name_keyword(parser, test_list):
    code = b"""
    call trace_entry(string="y", status=stat)
    call trace_exit(string="y", status=stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_wrong_name_by_param(parser, test_list):
    code = b"""
    character(len=100), parameter :: sub_name = "y"
    call trace_entry(sub_name, stat)
    call trace_exit(sub_name, stat)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_no_name(parser, test_list):
    code = b"""
    call trace_entry()
    call trace_exit()
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_unknown_name(parser, test_list):
    code = b"""
    call trace_entry(other_var)
    call trace_exit(other_var)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_correct_caps(parser, test_list):
    wrapped_code = b"""
    module foo
    SUBROUTINE X(Y)
    CALL TRACE_ENTRY("X", STAT)
    CALL TRACE_EXIT("X", STAT)
    END SUBROUTINE X
    end module foo
    """
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0, error_log.errors[0].message


def test_trace_entry_exit_correct_by_param_mixed_caps(parser, test_list):
    wrapped_code = b"""
    module foo
    SUBROUTINE X(Y)
    CHARACTER(len=100), PARAMETER :: sub_name = "X"
    CALL TRACE_ENTRY(SUB_NAME, STAT)
    CALL TRACE_EXIT(SUB_NAME, STAT)
    END SUBROUTINE X
    end module foo
    """
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_by_param_all_caps(parser, test_list):
    wrapped_code = b"""
    module foo
    SUBROUTINE X(Y)
    CHARACTER(len=100), PARAMETER :: SUB_NAME = "X"
    CALL TRACE_ENTRY(SUB_NAME, STAT)
    CALL TRACE_EXIT(SUB_NAME, STAT)
    END SUBROUTINE X
    end module foo
    """
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_function(parser, test_list):
    wrapped_code = b"""
    module foo
    function X(Y)
    CALL TRACE_ENTRY("x", STAT)
    CALL TRACE_EXIT("x", STAT)
    end function X
    end module foo
    """
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_missing_function(parser, test_list):
    wrapped_code = b"""
    module foo
    function X(Y)
    end function X
    end module foo
    """
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 2



"""Module to handle errors, warnings and info messages"""
from typing import ClassVar

from tree_sitter import Node


class FortranMsgBase:
    """Base class for other static analysis problems to inherit from"""

    ERROR_TYPE: ClassVar[str] = "NONE"
    ERROR_STYLE: ClassVar[str] = "grey"
    LINE_NUMBER_OFFSET = 8
    ERROR_SEVERITY: ClassVar[int] = 100

    def __init__(self, node: Node, message: str) -> None:
        self.message = message
        self.start_point = node.start_point
        self.end_point = node.end_point

    def print_err(self, filename: str, console) -> None:
        """Print the error to the supplied console"""
        console.print(self, style=self.ERROR_STYLE)
        context = self.context(filename, underline=True)
        if context:
            console.print(context)

    def context(self, filename, *, underline=False):
        """Print a line of context for the current error"""
        context = ""

        with open(filename, "rb") as fd:
            start_line, start_char = self.start_point
            end_line, end_char = self.end_point
            num_lines = end_line - start_line + 1
            num_chars = end_char - start_char

            file_str = str(filename)

            if num_lines == 1:
                line = fd.read().splitlines()[start_line].decode(errors="replace")
                context = f"{file_str}:{start_line+1:{self.LINE_NUMBER_OFFSET}}>{line}"
                if underline:
                    context += (
                        "\n"
                        + " " * (len(file_str) + 1)
                        + " " * (self.LINE_NUMBER_OFFSET + 1)
                        + " " * start_char
                        + "^" * num_chars
                    )
        return context

    def __repr__(self):
        return f"{self.ERROR_TYPE}: {self.message}"


class FortranError(FortranMsgBase):
    """Fatal static analysis problem in code"""

    ERROR_TYPE: ClassVar[str] = "Error"
    ERROR_STYLE: ClassVar[str] = "red"
    ERROR_SEVERITY: ClassVar[int] = 2


class FortranWarning(FortranMsgBase):
    """Warning message from static analysis"""

    ERROR_TYPE: ClassVar[str] = "Warning"
    ERROR_STYLE: ClassVar[str] = "yellow"
    ERROR_SEVERITY: ClassVar[int] = 1


class FortranInfo(FortranMsgBase):
    """Warning message from static analysis"""

    ERROR_TYPE: ClassVar[str] = "Info"
    ERROR_STYLE: ClassVar[str] = "Blue"
    ERROR_SEVERITY: ClassVar[int] = 0


def new_fortran_error(level: str, node: Node, message: str) -> FortranMsgBase:
    """Generate a new fortran diagnostic message"""
    cls = FortranMsgBase
    if level == "Error":
        cls = FortranError
    elif level == "Warning":
        cls = FortranWarning
    elif level == "Info":
        cls = FortranInfo
    else:
        raise ValueError("Unknown fortran diagnostic message type: " + level)
    return cls(node, message)


FORTRAN_ERRORS = {
    "Error": FortranError,
    "Warn": FortranWarning,
    "Info": FortranInfo,
}

ERROR_SEVERITY = {k: v.ERROR_SEVERITY for k, v in FORTRAN_ERRORS.items()}



"""Test that a subroutine or function has a trace_entry and trace_exit with the correct name"""
from tree_sitter import Node

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran import CallExpression, FType, VariableDeclaration, parser


@parser.node_type_check("subroutine", "function")
def test_trace_entry_exit(node: Node, error_log: ErrorLogger) -> None:
    """Test that a subroutine or function has a trace_entry and trace_exit with the correct name"""

    has_trace_entry = False
    has_trace_exit = False

    subroutine_name = node.get_child_property(parser.FORTRAN_CONTEXTS[node.type]).raw(lower=True)

    const_string_vars = {}

    for var_node in node.get_children_by_name("variable_declaration"):
        var_decl = VariableDeclaration(var_node)

        if var_decl.type != FType.CHARACTER:
            continue

        for var_name, initial_value in var_decl.vars.items():
            if initial_value:
                const_string_vars[var_name] = initial_value.lower()

    for statement in node.get_children_by_name("subroutine_call"):
        routine = CallExpression(statement)

        # routine = get_call_expression_name(statement)

        if routine.name == "trace_entry":
            has_trace_entry = True
        elif routine.name == "trace_exit":
            has_trace_exit = True

        if routine.name in ["trace_entry", "trace_exit"]:
            try:
                _, trace_node = routine.get_arg(position=1, keyword="string")
            except KeyError:
                err = f"Unparsable name passed to trace in {subroutine_name}"
                error_log.add_msg("Error", statement, err)
                continue

            if trace_node.type == "string_literal":
                trace_string = trace_node.parse_string_literal().lower()
                if trace_string != subroutine_name:
                    err = f"Incorrect name passed to trace in {subroutine_name}"
                    error_log.add_msg("Error", trace_node, err)

            elif trace_node.type == "identifier":
                trace_sub_text = trace_node.raw(lower=True)
                if trace_sub_text in const_string_vars:
                    trace_string = const_string_vars[trace_sub_text]
                    if trace_string.lower() != subroutine_name:
                        err = (
                            f"Incorrect name passed to trace in {subroutine_name} "
                            f'by variable {trace_sub_text}="{trace_string}"'
                        )
                        error_log.add_msg("Error", trace_node, err)
                else:
                    err = f"Unidentified variable {trace_sub_text} passed to trace in {subroutine_name}"
                    error_log.add_msg("Error", trace_node, err)

            else:
                err = f"Unrecognisable {statement.get_code()} {trace_node.type=} {statement}"
                raise ValueError(err)

    if not has_trace_entry:
        error_log.add_msg("Warning", node, f"Missing trace_entry in {subroutine_name}")
    if not has_trace_exit:
        error_log.add_msg("Warning", node, f"Missing trace_exit in {subroutine_name}")



"""Tests for Fortran code in CASTEP"""
from importlib import resources as impresources
from typing import Dict, List, Tuple

from tree_sitter import Language, Node, Parser

import castep_linter.fortran.monekey_patching as mp
from castep_linter.fortran.type_checking import node_type_check

# Monkey patch extra methods on node to get useful things like get_child_by_name
mp.add_extra_node_methods()


def get_fortran_parser():
    """Get a tree-sitter-fortran parser from src"""

    tree_sitter_src_ref = impresources.files("castep_linter") / "tree_sitter_fortran"
    with impresources.as_file(tree_sitter_src_ref) as tree_sitter_src:
        fortran_language = Language(tree_sitter_src / "fortran.so", "fortran")

    parser = Parser()
    parser.set_language(fortran_language)
    return parser


FORTRAN_CONTEXTS = {
    "subroutine": "subroutine_statement.name",
    "function": "function_statement.name",
    "module": "module_statement.name",
    "submodule": "submodule_statement.name",
    "program": "program_statement.name",
}


@node_type_check("argument_list")
def parse_arg_list(node: Node) -> Tuple[List[Node], Dict[str, Node]]:
    """
    Convert a fortran argument list into a args, kwargs pair.
    If lower_kwargs is true, the keyword arguments will be lowercased.
    """

    args = []
    kwargs = {}

    parsing_arg_list = True

    for child in node.children[1:-1:2]:
        if child.type == "keyword_argument":
            parsing_arg_list = False

        if parsing_arg_list:
            args.append(child)
        elif child.type == "keyword_argument":
            key, _, value = child.children
            kwargs[key.raw(lower=True)] = value

        else:
            err = f"Unknown argument list item in keyword arguments: {child.type}"
            raise ValueError(err)

    return args, kwargs



"""Module holding the CallExpression type"""
from typing import ClassVar, List, Optional, Tuple

from tree_sitter import Node

from castep_linter.fortran.argument_parser import ArgParser
from castep_linter.fortran.fortran_statement import FortranStatementParser


class CallExpression(FortranStatementParser):
    """Class representing a fortran call expression"""

    ALLOWED_NODES: ClassVar[List[str]] = ["call_expression", "subroutine_call"]

    def __init__(self, call_expression_node: Node) -> None:
        super().__init__(call_expression_node)

        self.name = self._get_name(call_expression_node)
        try:
            arg_list = call_expression_node.get_child_property("argument_list")
        except KeyError:
            self.args = []
        else:
            self.args = ArgParser(arg_list)

    def get_arg(self, keyword: str, position: Optional[int] = None) -> Tuple[ArgParser.ArgType, Node]:
        """Get an argument from the call expression"""
        return self.args.get(keyword, position)

    def _get_name(self, node):
        try:
            return node.get_child_property("identifier").raw(lower=True)
        except LookupError:
            return None

    def __str__(self):
        return f"{self.name=} {self.args=}"



""""Module with base class for Fortran code parser classes"""
from typing import ClassVar, List

from tree_sitter import Node

from castep_linter.fortran.type_checking import WrongNodeError, node_of_type


class FortranStatementParser:
    """Base class for fortran statement parsers"""

    ALLOWED_NODES: ClassVar[List[str]] = []

    def __init__(self, node: Node):
        if not node_of_type(node, self.ALLOWED_NODES):
            err = f"{node.type} not in {self.ALLOWED_NODES}"
            raise WrongNodeError(err)

        self.node = node



# pylint: disable=W0621,C0116,C0114
import pytest

from castep_linter import tests
from castep_linter.fortran.parser import get_fortran_parser
from castep_linter.scan_files import run_tests_on_code


@pytest.fixture
def test_list():
    return {"call_expression": [tests.test_allocate_has_stat]}


@pytest.fixture
def parser():
    return get_fortran_parser()


def subroutine_wrapper(code):
    return (
        b"""module foo
        subroutine x(y)
        """
        + code
        + b"""
        end subroutine x
        end module foo"""
    )


def test_allocate_stat_correct(parser, test_list):
    code = b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (u/=0) STOP 'err'
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_allocate_stat_correct_mixed_caps(parser, test_list):
    code = b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (U/=0) STOP 'err'
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0, error_log.errors[0].message


def test_allocate_stat_correct_mixed_caps2(parser, test_list):
    code = b"""
    allocate(stat_checked_var(x,y,z), stat=U)
    if (u/=0) STOP 'err'
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0, error_log.errors[0].message


def test_allocate_stat_correct_comment(parser, test_list):
    code = b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    ! comment
    if (u/=0) STOP 'err'
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_allocate_no_stat(parser, test_list):
    code = b"""
    allocate(stat_checked_var(x,y,z))
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_allocate_stat_not_checked(parser, test_list):
    code = b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    """
    wrapped_code = subroutine_wrapper(code)
    error_log = run_tests_on_code(parser, wrapped_code, test_list, "filename")
    assert len(error_log.errors) == 1



""""Tests to be performed by the CASTEP Fortran linter"""
from castep_linter.tests.allocate_stat_checked import test_allocate_has_stat
from castep_linter.tests.complex_has_dp import test_complex_has_dp
from castep_linter.tests.has_trace_entry_exit import test_trace_entry_exit
from castep_linter.tests.number_literal_correct_kind import test_number_literal
from castep_linter.tests.real_declaration_has_dp import test_real_dp_declaration

test_list = {
    "variable_declaration": [test_real_dp_declaration],
    "subroutine": [test_trace_entry_exit],
    "function": [test_trace_entry_exit],
    "call_expression": [test_complex_has_dp, test_allocate_has_stat],
    "number_literal": [test_number_literal],
}



# pylint: disable=W0621,C0116,C0114
import pytest

from castep_linter import tests
from castep_linter.fortran.parser import get_fortran_parser
from castep_linter.scan_files import run_tests_on_code


@pytest.fixture
def test_list():
    return {"number_literal": [tests.test_number_literal]}


@pytest.fixture
def parser():
    return get_fortran_parser()


def test_integer_literal_no_dp(parser, test_list):
    code = b"z = 1"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_integer_literal_has_dp(parser, test_list):
    code = b"z = 1_dp"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1



"""Fortran parsing helpers"""
# flake8: noqa
from castep_linter.fortran.argument_parser import ArgType
from castep_linter.fortran.call_expression import CallExpression
from castep_linter.fortran.variable_declaration import FType, VariableDeclaration



"""Test that a number literal has a dp (if real) or no dp if of any other type"""
from tree_sitter import Node

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran import parser


@parser.node_type_check("number_literal")
def test_number_literal(node: Node, error_log: ErrorLogger) -> None:
    """Test that a number literal has a dp (if real) or no dp if of any other type"""

    def is_int(x: str) -> bool:
        return not any(c in x.lower() for c in [".", "e", "d"])

    literal_string = node.raw(lower=True)

    if "_" in literal_string:
        value, kind = literal_string.split("_", maxsplit=1)

        if is_int(value):
            error_log.add_msg("Error", node, f"Integer literal with {kind=}")
        elif kind != "dp":
            error_log.add_msg("Error", node, f"Float literal with {kind=}")

    elif "d" in literal_string:
        pass  # eg 5.0d4
    elif not is_int(literal_string):
        error_log.add_msg("Error", node, "Float literal without kind")



"""Static code analysis tool for castep"""
import argparse
import pathlib
import sys
from typing import Generator

from rich.console import Console
from tree_sitter import Node, Parser, Tree

from castep_linter import error_logging
from castep_linter.error_logging.xml_writer import write_xml
from castep_linter.fortran import parser
from castep_linter.tests import test_list


def traverse_tree(tree: Tree) -> Generator[Node, None, None]:
    """Traverse a tree-sitter tree in a depth first search"""
    cursor = tree.walk()

    reached_root = False
    while not reached_root:
        yield cursor.node

        if cursor.goto_first_child():
            continue

        if cursor.goto_next_sibling():
            continue

        retracing = True
        while retracing:
            if not cursor.goto_parent():
                retracing = False
                reached_root = True

            if cursor.goto_next_sibling():
                retracing = False


# done - complex(var) vs complex(var,dp) or complex(var, kind=dp)
# done - allocate without stat and stat not checked. deallocate?
# done - integer_dp etc
# real with trailing . not .0 or .0_dp?
# io_allocate_abort with wrong subname
# tabs & DOS line endings, whitespace, comments?


def run_tests_on_code(fort_parser: Parser, code: bytes, test_dict: dict, filename: str) -> error_logging.ErrorLogger:
    """Run all available tests on the supplied source code"""
    tree = fort_parser.parse(code)
    error_log = error_logging.ErrorLogger(filename)

    for node in traverse_tree(tree):
        # Have to check for is_named here as we want the statements,
        # not literal words like subroutine
        if node.is_named and node.type in test_dict:
            for test in test_dict[node.type]:
                test(node, error_log)

    return error_log


def path(arg: str) -> pathlib.Path:
    """Check a file exists and if so, return a path object"""
    my_file = pathlib.Path(arg)
    if not my_file.is_file():
        err = f"The file {arg} does not exist!"
        raise argparse.ArgumentTypeError(err)
    return my_file


def parse_args():
    """Parse the command line args for a message print level and a list of filenames"""
    arg_parser = argparse.ArgumentParser(prog="castep-linter", description="Code linter for CASTEP")
    arg_parser.add_argument(
        "-l",
        "--level",
        help="Error message level",
        default="Info",
        choices=error_logging.ERROR_SEVERITY.keys(),
    )
    arg_parser.add_argument("-x", "--xml", type=pathlib.Path, help="File for JUnit xml output if required")
    arg_parser.add_argument("-q", "--quiet", action="store_true", help="Do not write to console")
    arg_parser.add_argument("file", nargs="+", type=path, help="Files to scan")
    return arg_parser.parse_args()


def main() -> None:
    """Main entry point for the CASTEP linter"""
    args = parse_args()

    fortran_parser = parser.get_fortran_parser()
    console = Console(soft_wrap=True)

    error_logs = {}

    for file in args.file:
        with file.open("rb") as fd:
            raw_text = fd.read()

        error_log = run_tests_on_code(fortran_parser, raw_text, test_list, str(file))

        if not args.quiet:
            error_log.print_errors(console, level=args.level)

            err_count = error_log.count_errors()

            console.print(
                f"{len(error_log.errors)} issues in {file} ({err_count['Error']} errors,"
                f" {err_count['Warn']} warnings, {err_count['Info']} info)"
            )

        error_logs[str(file)] = error_log

    if args.xml:
        write_xml(args.xml, error_logs, error_logging.ERROR_SEVERITY[args.level])

    if any(e.has_errors for e in error_logs.values()):
        sys.exit(1)
    else:
        sys.exit(0)



"""Module holding methods for type checking tree_sitter Node functions"""
import functools
from typing import List, Union

from tree_sitter import Node


class WrongNodeError(Exception):
    """Exception thrown when an invalid node is passed to a typed function"""


def node_type_check(*types: List[str]):
    """Check a node is of a certain type(s)"""

    def decorator_node_type_check(func):
        @functools.wraps(func)
        def wrapped_func(node: Node, *args, **kwargs):
            if not node.is_named:
                err = f"Wrong node type passed: {node.type} when {types} was expected"
                raise WrongNodeError(err)
            if not node_of_type(node, types):
                err = f"Wrong node type passed: unnamed ({node.type}) when {types} was expected"
                raise WrongNodeError(err)
            return func(node, *args, **kwargs)

        return wrapped_func

    return decorator_node_type_check


def node_of_type(node, typename: Union[str, List[str]]) -> bool:
    """Check if a node is of this type"""
    if isinstance(typename, str):
        return node.is_named and node.type == typename

    return node.is_named and any(node.type == t for t in typename)



"""Module holding the VariableDeclaration type"""
from enum import Enum, auto
from typing import ClassVar, Dict, List, Optional, Set, Tuple

from tree_sitter import Node

from castep_linter.fortran.argument_parser import ArgParser
from castep_linter.fortran.fortran_statement import FortranStatementParser


class FType(Enum):
    """Intrinsic variable types in fortran"""

    REAL = auto()
    DOUBLE = auto()
    COMPLEX = auto()
    INTEGER = auto()
    LOGICAL = auto()
    CHARACTER = auto()
    OTHER = auto()


def parse_fort_type(var_decl_node: Node) -> FType:
    """Parse a variable declaration for type"""
    try:
        fortran_type = var_decl_node.get_child_property("intrinsic_type").raw().upper()
        if fortran_type == "DOUBLE PRECISION":
            return FType.DOUBLE
        else:
            return FType[fortran_type]
    except KeyError:
        return FType.OTHER


def parse_fort_type_qualifiers(var_decl_node: Node) -> Set[str]:
    """Parse a variable declaration for qualifiers, eg parameter"""
    qualifiers = set()
    for type_qualifier in var_decl_node.get_children_by_name("type_qualifier"):
        qualifier = type_qualifier.raw(lower=True)
        qualifiers.add(qualifier)
    return qualifiers


def parse_fort_var_size(var_decl_node: Node) -> ArgParser:
    """Parse a variable declaration for a size, eg kind=8"""
    try:
        fortran_size = var_decl_node.get_child_property("size")
    except KeyError:
        return ArgParser()

    return ArgParser(fortran_size.get_child_property("argument_list"))


def parse_fort_var_names(var_decl_node: Node) -> Dict[str, Optional[str]]:
    """Parse variable declaration statement for variables and optionally assignments"""
    myvars: Dict[str, Optional[str]] = {}
    for assignment in var_decl_node.get_children_by_name("assignment_statement"):
        lhs, rhs = assignment.split()
        #   lhs, rhs = split_relational_node(assignment)
        varname = lhs.raw(lower=True)
        if rhs.type == "string_literal":
            myvars[varname] = rhs.parse_string_literal()
        else:
            myvars[varname] = None
    return myvars


class VariableDeclaration(FortranStatementParser):
    """Class representing a variable declaration"""

    ALLOWED_NODES: ClassVar[List[str]] = ["variable_declaration"]

    def __init__(self, var_decl_node: Node) -> None:
        super().__init__(var_decl_node)

        self.type = parse_fort_type(var_decl_node)
        self.qualifiers = parse_fort_type_qualifiers(var_decl_node)
        self.vars = parse_fort_var_names(var_decl_node)
        self.args = parse_fort_var_size(var_decl_node)

    def get_arg(self, keyword: str, position: Optional[int] = None) -> Tuple[ArgParser.ArgType, Node]:
        """Get an argument from the call expression"""
        return self.args.get(keyword, position)



# pylint: disable=W0621,C0116,C0114
import pytest

from castep_linter import tests
from castep_linter.fortran.parser import get_fortran_parser
from castep_linter.scan_files import run_tests_on_code


@pytest.fixture
def test_list():
    return {"call_expression": [tests.test_complex_has_dp]}


@pytest.fixture
def parser():
    return get_fortran_parser()


def test_other_function(parser, test_list):
    code = b"y = myownfunction(a, b, dp)"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_complex_dp_correct(parser, test_list):
    code = b"y = CMPLX(a, b, dp)"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_complex_dp_correct_keyword(parser, test_list):
    code = b"y = CMPLX(z, kind=dp)"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 0


def test_complex_dp_wrong_place(parser, test_list):
    code = b"y = CMPLX(z, dp)"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_complex_dp_missing(parser, test_list):
    code = b"y = CMPLX(a, b)"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1


def test_complex_wrong_kind(parser, test_list):
    code = b"y = CMPLX(a, b, x)"
    error_log = run_tests_on_code(parser, code, test_list, "filename")
    assert len(error_log.errors) == 1



"""Module containing an argument list parser"""
from enum import Enum, auto
from typing import ClassVar, List, Optional

from tree_sitter import Node

from castep_linter.fortran.parser import parse_arg_list


class ArgType(Enum):
    """Types of arguments for fortran functions/subroutines"""

    KEYWORD = auto()
    POSITION = auto()
    # NONE = auto()


class ArgParser:
    """Parser for fortran argument lists"""

    ALLOWED_NODES: ClassVar[List[str]] = ["argument_list"]
    ArgType = ArgType

    def __init__(self, arg_list: Optional[Node] = None):
        if arg_list:
            self.args, self.kwargs = parse_arg_list(arg_list)
        else:
            self.args, self.kwargs = [], {}

    def get(self, keyword: str, position: Optional[int] = None):
        """Return a value from a fortran argument list by keyword and optionally position"""
        if position and len(self.args) >= position:
            return ArgType.POSITION, self.args[position - 1]
        if keyword.lower() in self.kwargs:
            return ArgType.KEYWORD, self.kwargs[keyword]

        err = f"Argument {keyword} not found in argument list"
        raise KeyError(err)
